import	os
import	re
import	unittest
from	pygwarts.magical.philosophers_stone		import Transmutable
from	pygwarts.tests.irma						import IrmaTestCase
from	pygwarts.irma.contrib					import LibraryContrib
from	pygwarts.irma.access					import LibraryAccess
from	pygwarts.irma.access.volume				import LibraryVolume
from	pygwarts.irma.access.bookmarks			import VolumeBookmark
from	pygwarts.irma.access.handlers.counters	import AccessCounter
from	pygwarts.irma.access.handlers.parsers	import TargetHandler
from	pygwarts.irma.access.handlers.parsers	import TargetStringAccumulator
from	pygwarts.irma.access.inducers.counters	import HandlerCounterInducer
from	pygwarts.irma.access.inducers.counters	import RegisterCounterInducer
from	pygwarts.irma.access.inducers.recap		import RegisterRecapInducer
from	pygwarts.irma.access.inducers.recap		import RegisterRecapAccumulatorInducer
from	pygwarts.irma.access.annex				import VolumeAnnex
from	pygwarts.irma.access.annex				import LibraryAnnex
from	pygwarts.irma.access.utils				import TextWrapper








class AccessCase(IrmaTestCase):

	"""
		Testing for fields and edge cases
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.ACCESS_LIBRARY): os.remove(cls.ACCESS_LIBRARY)

		if	os.path.isfile(cls.majorV1): os.remove(cls.majorV1)
		if	os.path.isfile(cls.majorV2): os.remove(cls.majorV2)

	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.ACCESS_LIBRARY)
		cls.majorV1 = str(cls.IRMA_ROOT /"majorV1.txt")
		cls.majorV2 = str(cls.IRMA_ROOT /"majorV2.txt")
		cls.fmake(

			cls, cls.majorV1,
			"31/12/2024 2359 @MajorData WARNING : All unsaved data will be gone next year\n"
			"01/01/2025 0000 @MajorData INFO : The new year has come\n"
			"01/01/2025 0001 @MajorData INFO : The new year has come one mintue now\n"
			"01/01/2025 0100 @MajorData INFO : The new year has come one hour now\n"
			"01/01/2025 0200 @MajorData INFO : The new year has come two hours now\n"
			"01/01/2025 0200 @MajorData INFO : It is time to destroy past year data\n"
			"01/01/2025 0200 @MajorData INFO : Past year data is 2195390253890235472904\n"
			"01/01/2025 0200 @MajorData WARNING : Past year data can't be destroyed because it's bad\n"
		)
		cls.fmake(

			cls, cls.majorV2,
			"02/01/2025 0000 @MajorData INFO : The new day of the new year has come\n"
			"02/01/2025 0001 @MajorData INFO : The new day of the new year has come one mintue now\n"
			"02/01/2025 0100 @MajorData INFO : The new day of the new year has come one hour now\n"
			"02/01/2025 0200 @MajorData INFO : The new day of the new year has come two hours now\n"
			"02/01/2025 0200 @MajorData INFO : It is time to destroy past year data\n"
			"02/01/2025 0200 @MajorData INFO : Found past year user John\n"
			"02/01/2025 0200 @MajorData INFO : Found past year user David\n"
			"02/01/2025 0200 @MajorData INFO : Found past year user Mia\n"
			"02/01/2025 0200 @MajorData WARNING : Past year data can't be destroyed because it's still bad\n"
		)


	def test_no_library(self):

		class LonelyAccess(LibraryAccess):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "no_library"
				init_level	= 10

		with self.assertLogs("no_library", 10) as case_loggy : self.test_case = LonelyAccess()
		self.assertIn(f"DEBUG:no_library:Library access point created", case_loggy.output)




	def test_dummy_library(self):

		class DummyLibrary(LibraryAccess):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "dummy_library"
				init_level	= 10

			class CommonBookmark1(VolumeBookmark):		trigger = "trigger1"
			class CommonBookmark2(VolumeBookmark):		trigger = "trigger2"
			class TheVolumeOne(LibraryVolume):			pass
			class TheVolumeTwo(LibraryVolume):

				class VolumesBookmark1(VolumeBookmark):	trigger = "trigger3"
				class VolumesBookmark2(VolumeBookmark):	trigger = "trigger4"

		with self.assertLogs("dummy_library", 10) as case_loggy : self.test_case = DummyLibrary()
		self.assertIn(f"DEBUG:dummy_library:Assigned to library {self.test_case}", case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:dummy_library:Assigned to library {self.test_case}"),2
		)
		self.assertIn(
			f"DEBUG:dummy_library:Assigned to volume {self.test_case.TheVolumeTwo}", case_loggy.output
		)
		self.assertEqual(

			case_loggy.output.count(f"DEBUG:dummy_library:Assigned to volume {self.test_case.TheVolumeTwo}"),
			2
		)
		self.assertIn(f"DEBUG:dummy_library:Library access point created", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.CommonBookmark1], "bookmark")
		self.assertEqual(self.test_case[self.test_case.CommonBookmark2], "bookmark")
		self.assertEqual(

			self.test_case.keysof("bookmark"),
			[ self.test_case.CommonBookmark1,self.test_case.CommonBookmark2 ]
		)
		self.assertEqual(self.test_case[self.test_case.TheVolumeOne], "volume")
		self.assertEqual(self.test_case[self.test_case.TheVolumeTwo], "volume")
		self.assertEqual(
			self.test_case.keysof("volume"),[ self.test_case.TheVolumeOne,self.test_case.TheVolumeTwo ]
		)
		self.assertEqual(

			self.test_case.TheVolumeTwo.keysof("bookmark"),
			[ self.test_case.TheVolumeTwo.VolumesBookmark1,self.test_case.TheVolumeTwo.VolumesBookmark2 ]
		)
		self.assertEqual(len(self.test_case),4)




	def test_broke_dummy_library(self):

		class DummyLibrary(LibraryAccess):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "broke_dummy_library"
				init_level	= 10

			class CommonBookmark1(VolumeBookmark):		pass
			class CommonBookmark2(VolumeBookmark):		pass
			class TheVolumeOne(LibraryVolume):			pass
			class TheVolumeTwo(LibraryVolume):

				class VolumesBookmark1(VolumeBookmark):	pass
				class VolumesBookmark2(VolumeBookmark):	pass

		with self.assertLogs("broke_dummy_library", 10) as case_loggy : self.test_case = DummyLibrary()
		self.assertIn(f"DEBUG:broke_dummy_library:Assigned to {self.test_case}", case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:broke_dummy_library:Assigned to {self.test_case}"),2)
		self.assertIn(f"DEBUG:broke_dummy_library:Incomplete bookmark", case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:broke_dummy_library:Incomplete bookmark"),4)
		self.assertIn(f"DEBUG:broke_dummy_library:Library access point created", case_loggy.output)
		self.assertIsNone(self.test_case[str(self.test_case)])
		self.assertIsNone(self.test_case[str(self.test_case.TheVolumeTwo)])
		self.assertEqual(self.test_case[self.test_case.TheVolumeOne], "volume")
		self.assertEqual(self.test_case[self.test_case.TheVolumeTwo], "volume")
		self.assertEqual(len(self.test_case),2)
		self.assertEqual(
			self.test_case.keysof("volume"),[ self.test_case.TheVolumeOne,self.test_case.TheVolumeTwo ]
		)








	def test_compile_triggers(self):

		class DummyLibrary(LibraryAccess):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "compile_triggers"
				init_level	= 10

		self.test_case = DummyLibrary()
		triggers = [ "trigger #1", "trigger #2", "trigger #3" ]

		with self.assertLogs("compile_triggers", 10) as case_loggy:
			self.assertEqual(

				self.test_case.compile_triggers(triggers),
				re.compile("(trigger #1)|(trigger #2)|(trigger #3)")
			)
		self.assertIn("DEBUG:compile_triggers:Compiling trigger \"trigger #1\"", case_loggy.output)
		self.assertIn("DEBUG:compile_triggers:Compiling trigger \"trigger #2\"", case_loggy.output)
		self.assertIn("DEBUG:compile_triggers:Compiling trigger \"trigger #3\"", case_loggy.output)




	def test_compile_bad_triggers(self):

		class DummyLibrary(LibraryAccess):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "compile_bad_triggers"
				init_level	= 10

		self.test_case = DummyLibrary()
		for bad_triggers in (

			"triggers", 42, .69, True, False, None, ..., print, Transmutable,
			( "trigger #1", "trigger #2", "trigger #3" ),
			{ "trigger #1", "trigger #2", "trigger #3" },
			{ 1: "trigger #1", 2: "trigger #2", 3: "trigger #3" },
		):
			with self.subTest(triggers=bad_triggers):
				with self.assertLogs("compile_bad_triggers", 10) as case_loggy:

					self.assertIsNone(self.test_case.compile_triggers(bad_triggers))
				self.assertIn(
					f"DEBUG:compile_bad_triggers:Invalid triggers {type(bad_triggers)}", case_loggy.output
				)




	def test_compile_mixed_triggers(self):

		class DummyLibrary(LibraryAccess):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "compile_mixed_triggers"
				init_level	= 10

		self.test_case = DummyLibrary()
		triggers = [ "trigger #1",[ "trigger #2" ], 42 ]
		with self.assertLogs("compile_mixed_triggers", 10) as case_loggy:
			self.assertEqual(

				self.test_case.compile_triggers(triggers),
				re.compile("(trigger #1)")
			)
		self.assertIn("DEBUG:compile_mixed_triggers:Compiling trigger \"trigger #1\"", case_loggy.output)
		self.assertIn(

			"DEBUG:compile_mixed_triggers:Skipping invalid trigger \"['trigger #2']\"",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:compile_mixed_triggers:Skipping invalid trigger \"42\"",
			case_loggy.output
		)




	def test_compile_no_triggers(self):

		class DummyLibrary(LibraryAccess):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "compile_no_triggers"
				init_level	= 10

		self.test_case = DummyLibrary()
		triggers = [( "trigger #1", ),[ "trigger #2" ],{ "trigger #3" },42 ]
		with self.assertLogs("compile_no_triggers", 10) as case_loggy:
			self.assertIsNone(self.test_case.compile_triggers(triggers))

		self.assertIn(

			"DEBUG:compile_no_triggers:Skipping invalid trigger \"('trigger #1',)\"",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:compile_no_triggers:Skipping invalid trigger \"['trigger #2']\"",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:compile_no_triggers:Skipping invalid trigger \"{'trigger #3'}\"",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:compile_no_triggers:Skipping invalid trigger \"42\"",
			case_loggy.output
		)
		self.assertIn("DEBUG:compile_no_triggers:No triggers compiled", case_loggy.output)


		with self.assertLogs("compile_no_triggers", 10) as case_loggy:
			self.assertIsNone(self.test_case.compile_triggers([]))
		self.assertIn("DEBUG:compile_no_triggers:No triggers compiled", case_loggy.output)








	def test_major_library(self):
		class MajorLibrary(LibraryAccess):

			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "major_library"
				init_level	= 10

			@TextWrapper("Major Library:\n")
			class MajorAnnex(LibraryAnnex):						pass
			class INFOS(VolumeBookmark):

				trigger	= "INFO : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tINFOS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class WARNINGS(VolumeBookmark):

				trigger	= "WARNING : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tWARNINGS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class ERRORS(VolumeBookmark):

				trigger	= "ERROR : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tERRORS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class V1(LibraryVolume):

				location = self.majorV1
				inrange = "01/01"
				@TextWrapper("\tVolume 1:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					trigger = "Past year data is "
					class Target(TargetHandler):

						rpattern = r".+ (?P<target>\d+)$"
						@TextWrapper("\t\tdata found: ", "\n")
						class Inducer(RegisterRecapInducer):	pass

			class V2(LibraryVolume):

				location = self.majorV2
				inrange	= "02/01/2025"
				@TextWrapper("\tVolume 2:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					trigger = "Found past year user "
					class Users(TargetStringAccumulator):

						rpattern = r".+ (?P<target>\w+)"
						@TextWrapper("\t\tusers: ","\n")
						class Inducer(RegisterRecapAccumulatorInducer):	joint = ", "


		self.test_case = MajorLibrary()
		self.assertEqual(

			self.test_case.MajorAnnex(),
			"Major Library:\n"
			"\tVolume 1:\n"
			"\t\tINFOS: 6/14\n"
			"\t\tWARNINGS: 1/2\n"
			"\t\tdata found: 2195390253890235472904\n"
			"\tVolume 2:\n"
			"\t\tINFOS: 8/14\n"
			"\t\tWARNINGS: 1/2\n"
			"\t\tusers: John, David, Mia\n"
		)








	def test_major_library_no_ranges(self):
		class MajorLibrary(LibraryAccess):

			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "major_library_no_ranges"
				init_level	= 10

			@TextWrapper("Major Library:\n")
			class MajorAnnex(LibraryAnnex):						pass
			class INFOS(VolumeBookmark):

				trigger	= "INFO : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tINFOS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class WARNINGS(VolumeBookmark):

				trigger	= "WARNING : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tWARNINGS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class ERRORS(VolumeBookmark):

				trigger	= "ERROR : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tERRORS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class V1(LibraryVolume):

				location = self.majorV1
				@TextWrapper("\tVolume 1:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					trigger = "Past year data is "
					class Target(TargetHandler):

						rpattern = r".+ (?P<target>\d+)$"
						@TextWrapper("\t\tdata found: ", "\n")
						class Inducer(RegisterRecapInducer):	pass

			class V2(LibraryVolume):

				location = self.majorV2
				@TextWrapper("\tVolume 2:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					trigger = "Found past year user "
					class Users(TargetStringAccumulator):

						rpattern = r".+ (?P<target>\w+)"
						@TextWrapper("\t\tusers: ","\n")
						class Inducer(RegisterRecapAccumulatorInducer):	joint = ", "


		self.test_case = MajorLibrary()
		self.assertIsNone(self.test_case.MajorAnnex())








	def test_major_library_no_locations(self):
		class MajorLibrary(LibraryAccess):

			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "major_library_no_locations"
				init_level	= 10

			@TextWrapper("Major Library:\n")
			class MajorAnnex(LibraryAnnex):						pass
			class INFOS(VolumeBookmark):

				trigger	= "INFO : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tINFOS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class WARNINGS(VolumeBookmark):

				trigger	= "WARNING : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tWARNINGS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class ERRORS(VolumeBookmark):

				trigger	= "ERROR : "
				class Counter(AccessCounter):

					@TextWrapper("\t\tERRORS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class V1(LibraryVolume):

				inrange = "01/01"
				@TextWrapper("\tVolume 1:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					trigger = "Past year data is "
					class Target(TargetHandler):

						rpattern = r".+ (?P<target>\d+)$"
						@TextWrapper("\t\tdata found: ", "\n")
						class Inducer(RegisterRecapInducer):	pass

			class V2(LibraryVolume):

				inrange	= "02/01/2025"
				@TextWrapper("\tVolume 2:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					trigger = "Found past year user "
					class Users(TargetStringAccumulator):

						rpattern = r".+ (?P<target>\w+)"
						@TextWrapper("\t\tusers: ","\n")
						class Inducer(RegisterRecapAccumulatorInducer):	joint = ", "


		self.test_case = MajorLibrary()
		self.assertIsNone(self.test_case.MajorAnnex())








	def test_major_library_no_triggers(self):
		class MajorLibrary(LibraryAccess):

			class loggy(LibraryContrib):

				handler		= self.ACCESS_LIBRARY
				init_name	= "major_library_no_triggers"
				init_level	= 10

			@TextWrapper("Major Library:\n")
			class MajorAnnex(LibraryAnnex):						pass
			class INFOS(VolumeBookmark):

				class Counter(AccessCounter):

					@TextWrapper("\t\tINFOS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class WARNINGS(VolumeBookmark):

				class Counter(AccessCounter):

					@TextWrapper("\t\tWARNINGS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class ERRORS(VolumeBookmark):

				class Counter(AccessCounter):

					@TextWrapper("\t\tERRORS: ")
					class Current(RegisterCounterInducer):		pass
					@TextWrapper("/","\n")
					class Total(HandlerCounterInducer):			pass

			class V1(LibraryVolume):

				location = self.majorV1
				inrange = "01/01"
				@TextWrapper("\tVolume 1:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					class Target(TargetHandler):

						rpattern = r".+ (?P<target>\d+)$"
						@TextWrapper("\t\tdata found: ", "\n")
						class Inducer(RegisterRecapInducer):	pass

			class V2(LibraryVolume):

				location = self.majorV2
				inrange	= "02/01/2025"
				@TextWrapper("\tVolume 2:\n")
				class Annex(VolumeAnnex):						pass
				class B1(VolumeBookmark):

					class Users(TargetStringAccumulator):

						rpattern = r".+ (?P<target>\w+)"
						@TextWrapper("\t\tusers: ","\n")
						class Inducer(RegisterRecapAccumulatorInducer):	joint = ", "


		self.test_case = MajorLibrary()
		self.assertIsNone(self.test_case.MajorAnnex())








if	__name__ == "__main__" : unittest.main(verbosity=2)










from pathlib								import Path
from typing									import List
from typing									import Callable
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.hagrid.thrivables				import Tree
from pygwarts.hagrid.thrivables				import Copse








class Bloom(Transmutable):

	"""
		hagrid core functionality super class, that serves as a Tree object (target) dispatcher.
		Must be declared under the Tree or Copse object (literally for Copse object, means declaration of
		Tree objects must precede Bloom objects). In initiation time the "innervate" method will search the
		"_UPPER_LAYER" field and check it's type. If Bloom object was not declared under the correct object,
		it will discard it self and will not participate in Tree processing. Once attached to the right
		object, the "innervate" method will scan all reachable Copses and their Trees to include itself in
		every Tree, that means every affected Tree object will be processed by current Bloom object.
		The Tree object processing refers to child class __call__ implementation, that when invoked
		must dispose source folders and files to be processed against the target folder. During every
		Bloom object call corresponding arguments (lists of folders or/and files or/and source folder
		or/and target folder) must be processed by Bloom object __call__ itself, or, how it is originally
		designed, by special callable, and in both situations this whole operation is called "planting".
		Any "planting" that will occur inside Bloom object invocation must affect only single Tree object
		that it currently processing, and only corresponding source location. In order to do that,
		__call__ must implement special "bloom-interface":
			accepts:
				origin	- current Tree object to be processed;
				sprout	- current source root folder as a string;
				branch	- current source working directory as a Path object;
				twigs	- list of Path objects as current source folders;
				leafs	- list of Path objects as current source files.
			return:
				None.
	"""

	__call__	:Callable[[Tree, str, Path, List[Path], List[Path]], None]


	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


		self.blooming = dict()
		self.spread_count = int()


		if	hasattr(self, "_UPPER_LAYER"):


			self.innervate(self._UPPER_LAYER)
			self.loggy.debug(f"Spreading complete, number of affected trees: {self.spread_count}")
		else:
			self.loggy.warning(f"No starting point for {self} spreading")




	def innervate(self, origin :Copse | Tree):

		"""
			The method that when called in initiation time to recursively traverse through all reachable
			Copse objects, starting from the upper layer object current Bloom object attached to, to find
			all reachable Tree object and include itself in it. As every Tree object that met during
			"innervate" traverse must be Chest object child, to be included in the Tree object means
			that Chest object simply will be called with current Bloom object as argument and putted
			in the Chest. This process called "tree affection" and maintains counter "spread_count"
			that reflects the number of Tree object that was "affected" during innervate. The further
			Tree objects processing fully relies on this logic, that affected Tree objects contain
			in it's "chests" Bloom objects, that will do the thing. The Copse object that gets in the way
			of "innervate" is the point for recursive call to find Tree object it must contain.
		"""

		match origin:

			case Copse():
				for inner in origin:

					self.loggy.debug(f"Affecting copse {origin} inner {inner}")
					self.innervate(inner)


			case Tree():

				origin(self)
				self.spread_count += 1
				self.loggy.debug(f"{origin} tree affected")
			case _:
				self.loggy.warning(f"Invalid spread point \"{origin}\" for {self}")




	def mark_tree(self, tree_name :str, planter :str):

		"""
			Helper method that allows tracking the "planting state" and "bough validation" of every
			Tree object being processed by current Bloom object, by maintaining special dictionary
			"blooming". At first time for any Tree object successfully processed, which means
			successful "planting", that Tree object will be mapped in "blooming" with nested
			dictionary, that includes key-value pairs that serves as the marks of validated "planter"
			(working "planting" invocation) and validated "bough" (target root folder) for current
			Tree object being processed. In other words current helper method must be invoked only
			in case of success and will serves as a proof and as a guide for further "planting".
			For example when Bloom object tries to "plant" some "sprigs" (folders/files) with
			broken "planting" or invalid "bough", the original design for Bloom __call__
			implementation will announce only the first occurrence of such situation to be logged
			as warning message, logging any further occurrences as debug message, in order not
			to flood the log with the same problem. This announce restriction is done by altering
			the "blooming" dictionary after the first time the trouble is met, marking "planter"
			or/and "bough" keys with corresponding values. If such troubles will gone and after
			the very first successful "planting" that goes next, those "planter" and "bough" marks
			will be again set to be validated for current Tree object.
		"""

		if	self.blooming.get(tree_name) is None :		self.blooming[tree_name] = dict()
		if	not self.blooming[tree_name].get(planter):	self.blooming[tree_name][planter] = True
		if	not self.blooming[tree_name].get("bough"):	self.blooming[tree_name]["bough"] = True




	def announce_implement(self, tree_name :str, planter :str):

		"""
			Helper method that implements current Bloom object "planting" validation, by maintaining
			the "planter" key-value pair in "blooming" nested for current Tree object dictionary.
		"""

		if		self.blooming.get(tree_name) is None : self.blooming[tree_name] = dict()
		state = self.blooming[tree_name].get(planter)


		if		state or state is None:	level,self.blooming[tree_name][planter] = 40,False
		else:	level = 10


		self.loggy.log(level, f"{tree_name} doesn't implement {planter}")




	def announce_bough(self, tree_name :str, blooming :str):

		"""
			Helper method that implements current Tree object "bough" validation, by maintaining
			the "bough" key-value pair in "blooming" nested for current Tree object dictionary.
		"""

		if		self.blooming.get(tree_name) is None : self.blooming[tree_name] = dict()
		state = self.blooming[tree_name].get("bough")


		if		state or state is None:	level,self.blooming[tree_name]["bough"] = 40,False
		else:	level = 10


		self.loggy.log(level, f"Invalid tree {tree_name} or no bough to {blooming}")










import	os
import	unittest
from	random							import choice
from	pygwarts.tests.filch			import FilchTestCase
from	pygwarts.irma.contrib			import LibraryContrib
from	pygwarts.filch.marauders_map	import MaraudersMap








class MaraudersMapCases(FilchTestCase):

	"""
		Very simple tests
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.MARAUDERS_HANDLER): os.remove(cls.MARAUDERS_HANDLER)

		if	os.path.isfile(cls.CSV_MAP_1):	os.remove(cls.CSV_MAP_1)
		if	os.path.isfile(cls.CSV_MAP_2):	os.remove(cls.CSV_MAP_2)
		if	os.path.isfile(cls.CSV_MAP_3):	os.remove(cls.CSV_MAP_3)
		if	os.path.isfile(cls.CSV_MAP_4):	os.remove(cls.CSV_MAP_4)
		if	os.path.isfile(cls.CSV_MAP_5):	os.remove(cls.CSV_MAP_5)
		if	os.path.isfile(cls.CSV_MAP_6):	os.remove(cls.CSV_MAP_6)

	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.MARAUDERS_HANDLER)

		cls.CSV_MAP_1 = str(cls.FILCH_ROOT /"map1.csv")
		cls.CSV_MAP_2 = str(cls.FILCH_ROOT /"map2.csv")
		cls.CSV_MAP_3 = str(cls.FILCH_ROOT /"map3.csv")
		cls.CSV_MAP_4 = str(cls.FILCH_ROOT /"map4.csv")
		cls.CSV_MAP_5 = str(cls.FILCH_ROOT /"map5.csv")
		cls.CSV_MAP_6 = str(cls.FILCH_ROOT /"map6.csv")
		cls.ip0 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip1 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip2 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip3 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip4 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip5 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip6 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip7 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip8 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.ip9 = str(

			f"{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
			f".{choice(cls.valid_IP_chars)}.{choice(cls.valid_IP_chars)}"
		)
		cls.mac0 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac1 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac2 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac3 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac4 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac5 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac6 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac7 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac8 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.mac9 = str(

			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
			"-"
			f"{choice(cls.valid_MAC_chars)}"
			f"{choice(cls.valid_MAC_chars)}"
		)
		cls.lmac0 = cls.mac0.lower()
		cls.lmac1 = cls.mac1.lower()
		cls.lmac2 = cls.mac2.lower()
		cls.lmac3 = cls.mac3.lower()
		cls.lmac4 = cls.mac4.lower()
		cls.lmac5 = cls.mac5.lower()
		cls.lmac6 = cls.mac6.lower()
		cls.lmac7 = cls.mac7.lower()
		cls.lmac8 = cls.mac8.lower()
		cls.lmac9 = cls.mac9.lower()

		cls.name0 = "dev-lls-cata0"
		cls.name1 = "dev-lls-cata1"
		cls.name2 = "dev-lls-cata2"
		cls.name3 = "dev-lls-cata3"
		cls.name4 = "dev-lls-cata4"
		cls.name5 = "dev-lls-cata5"
		cls.name6 = "dev-lls-cata6"
		cls.name7 = "dev-lls-cata7"
		cls.name8 = "dev-lls-cata8"
		cls.name9 = "dev-lls-cata9"

		cls.desc0 = "room 0 switch"
		cls.desc1 = "room 1 switch"
		cls.desc2 = "room 2 switch"
		cls.desc3 = "room 3 switch"
		cls.desc4 = "room 4 switch"
		cls.desc5 = "room 5 switch"
		cls.desc6 = "room 6 switch"
		cls.desc7 = "room 7 switch"
		cls.desc8 = "room 8 switch"
		cls.desc9 = "room 9 switch"

		cls.fmake(

			cls,
			cls.CSV_MAP_1,
			f"{cls.ip0};{cls.mac0};{cls.name0};{cls.desc0}\n"
			f"{cls.ip1};{cls.mac1};{cls.name1};{cls.desc1}\n"
			f"{cls.ip2};{cls.mac2};{cls.name2};{cls.desc2}\n"
			f"{cls.ip3};{cls.mac3};{cls.name3};{cls.desc3}\n"
			f"{cls.ip4};{cls.mac4};{cls.name4};{cls.desc4}\n"
			f"{cls.ip5};{cls.mac5};{cls.name5};{cls.desc5}\n"
			f"{cls.ip6};{cls.mac6};{cls.name6};{cls.desc6}\n"
			f"{cls.ip7};{cls.mac7};{cls.name7};{cls.desc7}\n"
			f"{cls.ip8};{cls.mac8};{cls.name8};{cls.desc8}\n"
			f"{cls.ip9};{cls.mac9};{cls.name9};{cls.desc9}\n"
		)
		cls.fmake(

			cls,
			cls.CSV_MAP_2,
			"mac,name,ip,desc\n"
			f"{cls.mac0},{cls.name0},{cls.ip0},{cls.desc0}\n"
			f"{cls.mac1},{cls.name1},{cls.ip1},{cls.desc1}\n"
			f"{cls.mac2},{cls.name2},{cls.ip2},{cls.desc2}\n"
			f"{cls.mac3},{cls.name3},{cls.ip3},{cls.desc3}\n"
			f"{cls.mac4},{cls.name4},{cls.ip4},{cls.desc4}\n"
			f"{cls.mac5},{cls.name5},{cls.ip5},{cls.desc5}\n"
			f"{cls.mac6},{cls.name6},{cls.ip6},{cls.desc6}\n"
			f"{cls.mac7},{cls.name7},{cls.ip7},{cls.desc7}\n"
			f"{cls.mac8},{cls.name8},{cls.ip8},{cls.desc8}\n"
			f"{cls.mac9},{cls.name9},{cls.ip9},{cls.desc9}\n"
		)
		cls.fmake(

			cls,
			cls.CSV_MAP_3,
			f"{cls.mac0}\t{cls.ip0}\n"
			f"{cls.mac1}\t{cls.ip1}\n"
			f"{cls.mac2}\t{cls.ip2}\n"
			f"{cls.mac3}\t{cls.ip3}\n"
			f"{cls.mac4}\t{cls.ip4}\n"
			f"{cls.mac5}\t{cls.ip5}\n"
			f"{cls.mac6}\t{cls.ip6}\n"
			f"{cls.mac7}\t{cls.ip7}\n"
			f"{cls.mac8}\t{cls.ip8}\n"
			f"{cls.mac9}\t{cls.ip9}\n"
		)
		cls.fmake(

			cls,
			cls.CSV_MAP_4,
			f"{cls.ip0};{cls.mac0};{cls.name0};{cls.desc0}\n"
			f"{cls.ip1};{cls.mac1};{cls.name1};{cls.desc1}\n"
			f"{cls.ip2};{cls.mac2};{cls.name2};{cls.desc2}\n"
		)
		cls.fmake(

			cls,
			cls.CSV_MAP_5,
			"mac,name,ip,desc\n"
			f"{cls.mac3},{cls.name3},{cls.ip3},{cls.desc3}\n"
			f"{cls.mac4},{cls.name4},{cls.ip4},{cls.desc4}\n"
			f"{cls.mac5},{cls.name5},{cls.ip5},{cls.desc5}\n"
			f"{cls.mac6},{cls.name6},{cls.ip6},{cls.desc6}\n"
		)
		cls.fmake(

			cls,
			cls.CSV_MAP_6,
			f"{cls.mac7}\t{cls.ip7}\n"
			f"{cls.mac8}\t{cls.ip8}\n"
			f"{cls.mac9}\t{cls.ip9}\n"
		)


	def test_CSV_1_valid_load(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_1_valid_load",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_1_valid_load", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, MAC=1, NAME=2, DESC=3)

		self.assertIn("DEBUG:CSV_1_valid_load:10 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_1_valid_load:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_1_valid_load:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_1_valid_load:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_1_valid_load:10 MAC mappings total", case_loggy.output)
		self.assertEqual(

			self.test_case["IP4"][self.ip0],
			{
				"MAC":	self.mac0.lower(),
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip1],
			{
				"MAC":	self.mac1.lower(),
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip2],
			{
				"MAC":	self.mac2.lower(),
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip3],
			{
				"MAC":	self.mac3.lower(),
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip4],
			{
				"MAC":	self.mac4.lower(),
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip5],
			{
				"MAC":	self.mac5.lower(),
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip6],
			{
				"MAC":	self.mac6.lower(),
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip7],
			{
				"MAC":	self.mac7.lower(),
				"NAME":	self.name7,
				"DESC":	self.desc7,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip8],
			{
				"MAC":	self.mac8.lower(),
				"NAME":	self.name8,
				"DESC":	self.desc8,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip9],
			{
				"MAC":	self.mac9.lower(),
				"NAME":	self.name9,
				"DESC":	self.desc9,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac0],
			{
				"IP4":	self.ip0,
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac1],
			{
				"IP4":	self.ip1,
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac2],
			{
				"IP4":	self.ip2,
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac3],
			{
				"IP4":	self.ip3,
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac4],
			{
				"IP4":	self.ip4,
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac5],
			{
				"IP4":	self.ip5,
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac6],
			{
				"IP4":	self.ip6,
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac7],
			{
				"IP4":	self.ip7,
				"NAME":	self.name7,
				"DESC":	self.desc7,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac8],
			{
				"IP4":	self.ip8,
				"NAME":	self.name8,
				"DESC":	self.desc8,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac9],
			{
				"IP4":	self.ip9,
				"NAME":	self.name9,
				"DESC":	self.desc9,
			}
		)




	def test_CSV_2_valid_load(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_2_valid_load",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_2_valid_load", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_2, IP4=2, MAC=0, NAME=1, DESC=3)

		self.assertIn("DEBUG:CSV_2_valid_load:11 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_2_valid_load:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_2_valid_load:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_2_valid_load:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_2_valid_load:10 MAC mappings total", case_loggy.output)
		self.assertEqual(

			self.test_case["IP4"][self.ip0],
			{
				"MAC":	self.lmac0,
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip1],
			{
				"MAC":	self.lmac1,
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip2],
			{
				"MAC":	self.lmac2,
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip3],
			{
				"MAC":	self.lmac3,
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip4],
			{
				"MAC":	self.lmac4,
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip5],
			{
				"MAC":	self.lmac5,
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip6],
			{
				"MAC":	self.lmac6,
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip7],
			{
				"MAC":	self.lmac7,
				"NAME":	self.name7,
				"DESC":	self.desc7,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip8],
			{
				"MAC":	self.lmac8,
				"NAME":	self.name8,
				"DESC":	self.desc8,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip9],
			{
				"MAC":	self.lmac9,
				"NAME":	self.name9,
				"DESC":	self.desc9,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac0],
			{
				"IP4":	self.ip0,
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac1],
			{
				"IP4":	self.ip1,
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac2],
			{
				"IP4":	self.ip2,
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac3],
			{
				"IP4":	self.ip3,
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac4],
			{
				"IP4":	self.ip4,
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac5],
			{
				"IP4":	self.ip5,
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac6],
			{
				"IP4":	self.ip6,
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac7],
			{
				"IP4":	self.ip7,
				"NAME":	self.name7,
				"DESC":	self.desc7,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac8],
			{
				"IP4":	self.ip8,
				"NAME":	self.name8,
				"DESC":	self.desc8,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac9],
			{
				"IP4":	self.ip9,
				"NAME":	self.name9,
				"DESC":	self.desc9,
			}
		)




	def test_CSV_3_valid_load(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_3_valid_load",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_3_valid_load", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_3, "\t", IP4=1, MAC=0)

		self.assertIn("DEBUG:CSV_3_valid_load:10 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_3_valid_load:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_3_valid_load:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_3_valid_load:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_3_valid_load:10 MAC mappings total", case_loggy.output)
		self.assertEqual(

			self.test_case["IP4"][self.ip0],
			{
				"MAC":	self.lmac0,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip1],
			{
				"MAC":	self.lmac1,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip2],
			{
				"MAC":	self.lmac2,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip3],
			{
				"MAC":	self.lmac3,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip4],
			{
				"MAC":	self.lmac4,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip5],
			{
				"MAC":	self.lmac5,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip6],
			{
				"MAC":	self.lmac6,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip7],
			{
				"MAC":	self.lmac7,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip8],
			{
				"MAC":	self.lmac8,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip9],
			{
				"MAC":	self.lmac9,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac0],
			{
				"IP4":	self.ip0,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac1],
			{
				"IP4":	self.ip1,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac2],
			{
				"IP4":	self.ip2,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac3],
			{
				"IP4":	self.ip3,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac4],
			{
				"IP4":	self.ip4,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac5],
			{
				"IP4":	self.ip5,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac6],
			{
				"IP4":	self.ip6,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac7],
			{
				"IP4":	self.ip7,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac8],
			{
				"IP4":	self.ip8,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac9],
			{
				"IP4":	self.ip9,
				"NAME":	None,
				"DESC":	None,
			}
		)




	def test_CSV_1_invalid(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_1_invalid",
				init_level=10,
			)
		)

		with self.assertLogs("CSV_1_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, IP4=0, MAC=1, NAME=2, DESC=3)

		self.assertIn("DEBUG:CSV_1_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_1_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=3, MAC=3, NAME=3, DESC=3)

		self.assertIn("DEBUG:CSV_1_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_1_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=30)

		self.assertIn("DEBUG:CSV_1_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_1_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";")

		self.assertIn("DEBUG:CSV_1_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)




	def test_CSV_2_invalid(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_2_invalid",
				init_level=10,
			)
		)

		with self.assertLogs("CSV_2_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_2, "\t", IP4=2, MAC=0, NAME=1, DESC=3)

		self.assertIn("DEBUG:CSV_2_invalid:11 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_2_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_2, IP4=3, MAC=3, NAME=3, DESC=3)

		self.assertIn("DEBUG:CSV_2_invalid:11 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_2_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_2, IP4=30, MAC=30, NAME=30, DESC=30)

		self.assertIn("DEBUG:CSV_2_invalid:11 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_2_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_2)

		self.assertIn("DEBUG:CSV_2_invalid:11 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)




	def test_CSV_3_invalid(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_3_invalid",
				init_level=10,
			)
		)

		with self.assertLogs("CSV_3_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_3, IP4=1, MAC=0)

		self.assertIn("DEBUG:CSV_3_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_3_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_3, "\t", IP4=0, MAC=1)

		self.assertIn("DEBUG:CSV_3_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_3_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_3, "\t", IP4=10, MAC=10)

		self.assertIn("DEBUG:CSV_3_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)

		with self.assertLogs("CSV_3_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_3, "\t")

		self.assertIn("DEBUG:CSV_3_invalid:10 records read", case_loggy.output)
		self.assertEqual(len(self.test_case), 0)




	def test_CSV_onload(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_onload",
				init_level=10,
			)
		)

		with self.assertLogs("CSV_onload", 10) as case_loggy:

			self.test_case.CSV(self.CSV_MAP_4, ";", IP4=0, MAC=1, NAME=2, DESC=3)
			self.test_case.CSV(self.CSV_MAP_5, IP4=2, MAC=0, NAME=1, DESC=3)
			self.test_case.CSV(self.CSV_MAP_6, "\t", IP4=1, MAC=0)

		self.assertIn("DEBUG:CSV_onload:3 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:5 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:3 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:3 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:4 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:4 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:3 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:3 MAC mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:7 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:7 MAC mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_onload:10 MAC mappings total", case_loggy.output)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_onload:3 records read"),2)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_onload:3 new IP4 mappings done"),2)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_onload:3 new MAC mappings done"),2)
		self.assertEqual(

			self.test_case["IP4"][self.ip0],
			{
				"MAC":	self.lmac0,
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip1],
			{
				"MAC":	self.lmac1,
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip2],
			{
				"MAC":	self.lmac2,
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip3],
			{
				"MAC":	self.lmac3,
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip4],
			{
				"MAC":	self.lmac4,
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip5],
			{
				"MAC":	self.lmac5,
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip6],
			{
				"MAC":	self.lmac6,
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip7],
			{
				"MAC":	self.lmac7,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip8],
			{
				"MAC":	self.lmac8,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip9],
			{
				"MAC":	self.lmac9,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac0],
			{
				"IP4":	self.ip0,
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac1],
			{
				"IP4":	self.ip1,
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac2],
			{
				"IP4":	self.ip2,
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac3],
			{
				"IP4":	self.ip3,
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac4],
			{
				"IP4":	self.ip4,
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac5],
			{
				"IP4":	self.ip5,
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac6],
			{
				"IP4":	self.ip6,
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac7],
			{
				"IP4":	self.ip7,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac8],
			{
				"IP4":	self.ip8,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac9],
			{
				"IP4":	self.ip9,
				"NAME":	None,
				"DESC":	None,
			}
		)








	def test_CSV_reload(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_reload",
				init_level=10,
			)
		)

		with self.assertLogs("CSV_reload", 10) as case_loggy:

			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, MAC=1, NAME=2, DESC=3)
			self.test_case.CSV(self.CSV_MAP_2, IP4=2, MAC=0, NAME=1, DESC=3)
			self.test_case.CSV(self.CSV_MAP_3, "\t", IP4=1, MAC=0)

		self.assertIn("DEBUG:CSV_reload:10 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_reload:11 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_reload:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_reload:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_reload:0 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_reload:0 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_reload:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:CSV_reload:10 MAC mappings total", case_loggy.output)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_reload:10 records read"),2)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_reload:0 new IP4 mappings done"),2)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_reload:0 new MAC mappings done"),2)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_reload:10 IP4 mappings total"),3)
		self.assertEqual(case_loggy.output.count("DEBUG:CSV_reload:10 MAC mappings total"),3)
		self.assertEqual(

			self.test_case["IP4"][self.ip0],
			{
				"MAC":	self.lmac0,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip1],
			{
				"MAC":	self.lmac1,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip2],
			{
				"MAC":	self.lmac2,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip3],
			{
				"MAC":	self.lmac3,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip4],
			{
				"MAC":	self.lmac4,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip5],
			{
				"MAC":	self.lmac5,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip6],
			{
				"MAC":	self.lmac6,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip7],
			{
				"MAC":	self.lmac7,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip8],
			{
				"MAC":	self.lmac8,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip9],
			{
				"MAC":	self.lmac9,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac0],
			{
				"IP4":	self.ip0,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac1],
			{
				"IP4":	self.ip1,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac2],
			{
				"IP4":	self.ip2,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac3],
			{
				"IP4":	self.ip3,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac4],
			{
				"IP4":	self.ip4,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac5],
			{
				"IP4":	self.ip5,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac6],
			{
				"IP4":	self.ip6,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac7],
			{
				"IP4":	self.ip7,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac8],
			{
				"IP4":	self.ip8,
				"NAME":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac9],
			{
				"IP4":	self.ip9,
				"NAME":	None,
				"DESC":	None,
			}
		)








	def test_CSV_partial_load_1(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_partial_load_1",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_partial_load_1", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, NAME=2)

		self.assertIn("DEBUG:CSV_partial_load_1:10 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_1:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_1:10 IP4 mappings total", case_loggy.output)
		self.assertIsNone(self.test_case["MAC"])
		self.assertEqual(

			self.test_case["IP4"][self.ip0],
			{
				"NAME":	self.name0,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip1],
			{
				"NAME":	self.name1,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip2],
			{
				"NAME":	self.name2,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip3],
			{
				"NAME":	self.name3,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip4],
			{
				"NAME":	self.name4,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip5],
			{
				"NAME":	self.name5,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip6],
			{
				"NAME":	self.name6,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip7],
			{
				"NAME":	self.name7,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip8],
			{
				"NAME":	self.name8,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip9],
			{
				"NAME":	self.name9,
				"MAC":	None,
				"DESC":	None,
			}
		)








	def test_CSV_partial_load_2(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_partial_load_2",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_partial_load_2", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", MAC=1, DESC=3)

		self.assertIn("DEBUG:CSV_partial_load_2:10 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_2:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_2:10 MAC mappings total", case_loggy.output)
		self.assertIsNone(self.test_case["IP4"])
		self.assertEqual(

			self.test_case["MAC"][self.lmac0],
			{
				"DESC":	self.desc0,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac1],
			{
				"DESC":	self.desc1,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac2],
			{
				"DESC":	self.desc2,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac3],
			{
				"DESC":	self.desc3,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac4],
			{
				"DESC":	self.desc4,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac5],
			{
				"DESC":	self.desc5,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac6],
			{
				"DESC":	self.desc6,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac7],
			{
				"DESC":	self.desc7,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac8],
			{
				"DESC":	self.desc8,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac9],
			{
				"DESC":	self.desc9,
				"IP4":	None,
				"NAME":	None,
			}
		)








	def test_CSV_partial_load_3(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_partial_load_4",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_partial_load_4", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0)

		self.assertIn("DEBUG:CSV_partial_load_4:10 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_4:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_4:10 IP4 mappings total", case_loggy.output)
		self.assertIsNone(self.test_case["MAC"])
		self.assertEqual(

			self.test_case["IP4"][self.ip0],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip1],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip2],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip3],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip4],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip5],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip6],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip7],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip8],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)
		self.assertEqual(

			self.test_case["IP4"][self.ip9],
			{
				"NAME":	None,
				"MAC":	None,
				"DESC":	None,
			}
		)








	def test_CSV_partial_load_4(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_partial_load_4",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_partial_load_4", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", MAC=1)

		self.assertIn("DEBUG:CSV_partial_load_4:10 records read", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_4:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:CSV_partial_load_4:10 MAC mappings total", case_loggy.output)
		self.assertIsNone(self.test_case["IP4"])
		self.assertEqual(

			self.test_case["MAC"][self.lmac0],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac1],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac2],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac3],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac4],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac5],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac6],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac7],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac8],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)
		self.assertEqual(

			self.test_case["MAC"][self.lmac9],
			{
				"DESC":	None,
				"IP4":	None,
				"NAME":	None,
			}
		)








	def test_CSV_partial_load_5(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="CSV_partial_load_3",
				init_level=10,
			)
		)
		with self.assertLogs("CSV_partial_load_3", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", NAME=2, DESC=3)

		self.assertIn("DEBUG:CSV_partial_load_3:10 records read", case_loggy.output)
		self.assertIsNone(self.test_case["IP4"])
		self.assertIsNone(self.test_case["MAC"])








	def test_mapping_properties(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="ip4_mapping_valid",
				init_level=10,
			)
		)
		with self.assertLogs("ip4_mapping_valid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, MAC=1, NAME=2, DESC=3)

		self.assertIn("DEBUG:ip4_mapping_valid:10 records read", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 MAC mappings total", case_loggy.output)

		self.assertEqual(len(self.test_case.ip4), 10)
		self.assertEqual(len(self.test_case.mac), 10)
		self.assertEqual(
			self.test_case.ip4,
			{
				self.ip0: {

					"MAC":	self.lmac0,
					"NAME":	self.name0,
					"DESC":	self.desc0,
				},
				self.ip1: {

					"MAC":	self.lmac1,
					"NAME":	self.name1,
					"DESC":	self.desc1,
				},
				self.ip2: {

					"MAC":	self.lmac2,
					"NAME":	self.name2,
					"DESC":	self.desc2,
				},
				self.ip3: {

					"MAC":	self.lmac3,
					"NAME":	self.name3,
					"DESC":	self.desc3,
				},
				self.ip4: {

					"MAC":	self.lmac4,
					"NAME":	self.name4,
					"DESC":	self.desc4,
				},
				self.ip5: {

					"MAC":	self.lmac5,
					"NAME":	self.name5,
					"DESC":	self.desc5,
				},
				self.ip6: {

					"MAC":	self.lmac6,
					"NAME":	self.name6,
					"DESC":	self.desc6,
				},
				self.ip7: {

					"MAC":	self.lmac7,
					"NAME":	self.name7,
					"DESC":	self.desc7,
				},
				self.ip8: {

					"MAC":	self.lmac8,
					"NAME":	self.name8,
					"DESC":	self.desc8,
				},
				self.ip9: {

					"MAC":	self.lmac9,
					"NAME":	self.name9,
					"DESC":	self.desc9,
				},
			}
		)
		self.assertEqual(
			list(self.test_case.ip4.keys()),
			[
				self.ip0,
				self.ip1,
				self.ip2,
				self.ip3,
				self.ip4,
				self.ip5,
				self.ip6,
				self.ip7,
				self.ip8,
				self.ip9
			]
		)
		self.assertEqual(
			list(self.test_case.ip4.values()),
			[
				{

					"MAC":	self.lmac0,
					"NAME":	self.name0,
					"DESC":	self.desc0,
				},
				{

					"MAC":	self.lmac1,
					"NAME":	self.name1,
					"DESC":	self.desc1,
				},
				{

					"MAC":	self.lmac2,
					"NAME":	self.name2,
					"DESC":	self.desc2,
				},
				{

					"MAC":	self.lmac3,
					"NAME":	self.name3,
					"DESC":	self.desc3,
				},
				{

					"MAC":	self.lmac4,
					"NAME":	self.name4,
					"DESC":	self.desc4,
				},
				{

					"MAC":	self.lmac5,
					"NAME":	self.name5,
					"DESC":	self.desc5,
				},
				{

					"MAC":	self.lmac6,
					"NAME":	self.name6,
					"DESC":	self.desc6,
				},
				{

					"MAC":	self.lmac7,
					"NAME":	self.name7,
					"DESC":	self.desc7,
				},
				{

					"MAC":	self.lmac8,
					"NAME":	self.name8,
					"DESC":	self.desc8,
				},
				{

					"MAC":	self.lmac9,
					"NAME":	self.name9,
					"DESC":	self.desc9,
				},
			]
		)
		self.assertEqual(
			self.test_case.mac,
			{
				self.lmac0: {

					"IP4":	self.ip0,
					"NAME":	self.name0,
					"DESC":	self.desc0,
				},
				self.lmac1: {

					"IP4":	self.ip1,
					"NAME":	self.name1,
					"DESC":	self.desc1,
				},
				self.lmac2: {

					"IP4":	self.ip2,
					"NAME":	self.name2,
					"DESC":	self.desc2,
				},
				self.lmac3: {

					"IP4":	self.ip3,
					"NAME":	self.name3,
					"DESC":	self.desc3,
				},
				self.lmac4: {

					"IP4":	self.ip4,
					"NAME":	self.name4,
					"DESC":	self.desc4,
				},
				self.lmac5: {

					"IP4":	self.ip5,
					"NAME":	self.name5,
					"DESC":	self.desc5,
				},
				self.lmac6: {

					"IP4":	self.ip6,
					"NAME":	self.name6,
					"DESC":	self.desc6,
				},
				self.lmac7: {

					"IP4":	self.ip7,
					"NAME":	self.name7,
					"DESC":	self.desc7,
				},
				self.lmac8: {

					"IP4":	self.ip8,
					"NAME":	self.name8,
					"DESC":	self.desc8,
				},
				self.lmac9: {

					"IP4":	self.ip9,
					"NAME":	self.name9,
					"DESC":	self.desc9,
				},
			}
		)
		self.assertEqual(
			list(self.test_case.mac.keys()),
			[
				self.lmac0,
				self.lmac1,
				self.lmac2,
				self.lmac3,
				self.lmac4,
				self.lmac5,
				self.lmac6,
				self.lmac7,
				self.lmac8,
				self.lmac9
			]
		)
		self.assertEqual(
			list(self.test_case.mac.values()),
			[
				{

					"IP4":	self.ip0,
					"NAME":	self.name0,
					"DESC":	self.desc0,
				},
				{

					"IP4":	self.ip1,
					"NAME":	self.name1,
					"DESC":	self.desc1,
				},
				{

					"IP4":	self.ip2,
					"NAME":	self.name2,
					"DESC":	self.desc2,
				},
				{

					"IP4":	self.ip3,
					"NAME":	self.name3,
					"DESC":	self.desc3,
				},
				{

					"IP4":	self.ip4,
					"NAME":	self.name4,
					"DESC":	self.desc4,
				},
				{

					"IP4":	self.ip5,
					"NAME":	self.name5,
					"DESC":	self.desc5,
				},
				{

					"IP4":	self.ip6,
					"NAME":	self.name6,
					"DESC":	self.desc6,
				},
				{

					"IP4":	self.ip7,
					"NAME":	self.name7,
					"DESC":	self.desc7,
				},
				{

					"IP4":	self.ip8,
					"NAME":	self.name8,
					"DESC":	self.desc8,
				},
				{

					"IP4":	self.ip9,
					"NAME":	self.name9,
					"DESC":	self.desc9,
				},
			]
		)








	def test_ip4_mapping_valid(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="ip4_mapping_valid",
				init_level=10,
			)
		)
		with self.assertLogs("ip4_mapping_valid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, MAC=1, NAME=2, DESC=3)

		self.assertIn("DEBUG:ip4_mapping_valid:10 records read", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_valid:10 MAC mappings total", case_loggy.output)

		self.assertEqual(self.test_case.ip4map_mac(self.ip0), self.lmac0)
		self.assertEqual(self.test_case.ip4map_mac(self.ip1), self.lmac1)
		self.assertEqual(self.test_case.ip4map_mac(self.ip2), self.lmac2)
		self.assertEqual(self.test_case.ip4map_mac(self.ip3), self.lmac3)
		self.assertEqual(self.test_case.ip4map_mac(self.ip4), self.lmac4)
		self.assertEqual(self.test_case.ip4map_mac(self.ip5), self.lmac5)
		self.assertEqual(self.test_case.ip4map_mac(self.ip6), self.lmac6)
		self.assertEqual(self.test_case.ip4map_mac(self.ip7), self.lmac7)
		self.assertEqual(self.test_case.ip4map_mac(self.ip8), self.lmac8)
		self.assertEqual(self.test_case.ip4map_mac(self.ip9), self.lmac9)

		self.assertEqual(self.test_case.ip4map_name(self.ip0), self.name0)
		self.assertEqual(self.test_case.ip4map_name(self.ip1), self.name1)
		self.assertEqual(self.test_case.ip4map_name(self.ip2), self.name2)
		self.assertEqual(self.test_case.ip4map_name(self.ip3), self.name3)
		self.assertEqual(self.test_case.ip4map_name(self.ip4), self.name4)
		self.assertEqual(self.test_case.ip4map_name(self.ip5), self.name5)
		self.assertEqual(self.test_case.ip4map_name(self.ip6), self.name6)
		self.assertEqual(self.test_case.ip4map_name(self.ip7), self.name7)
		self.assertEqual(self.test_case.ip4map_name(self.ip8), self.name8)
		self.assertEqual(self.test_case.ip4map_name(self.ip9), self.name9)

		self.assertEqual(self.test_case.ip4map_desc(self.ip0), self.desc0)
		self.assertEqual(self.test_case.ip4map_desc(self.ip1), self.desc1)
		self.assertEqual(self.test_case.ip4map_desc(self.ip2), self.desc2)
		self.assertEqual(self.test_case.ip4map_desc(self.ip3), self.desc3)
		self.assertEqual(self.test_case.ip4map_desc(self.ip4), self.desc4)
		self.assertEqual(self.test_case.ip4map_desc(self.ip5), self.desc5)
		self.assertEqual(self.test_case.ip4map_desc(self.ip6), self.desc6)
		self.assertEqual(self.test_case.ip4map_desc(self.ip7), self.desc7)
		self.assertEqual(self.test_case.ip4map_desc(self.ip8), self.desc8)
		self.assertEqual(self.test_case.ip4map_desc(self.ip9), self.desc9)
		self.assertEqual(

			self.test_case.ip4map(self.ip0),
			{
				"MAC":	self.lmac0,
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip1),
			{
				"MAC":	self.lmac1,
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip2),
			{
				"MAC":	self.lmac2,
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip3),
			{
				"MAC":	self.lmac3,
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip4),
			{
				"MAC":	self.lmac4,
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip5),
			{
				"MAC":	self.lmac5,
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip6),
			{
				"MAC":	self.lmac6,
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip7),
			{
				"MAC":	self.lmac7,
				"NAME":	self.name7,
				"DESC":	self.desc7,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip8),
			{
				"MAC":	self.lmac8,
				"NAME":	self.name8,
				"DESC":	self.desc8,
			}
		)
		self.assertEqual(

			self.test_case.ip4map(self.ip9),
			{
				"MAC":	self.lmac9,
				"NAME":	self.name9,
				"DESC":	self.desc9,
			}
		)








	def test_ip4_mapping_invalid(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="ip4_mapping_invalid",
				init_level=10,
			)
		)
		with self.assertLogs("ip4_mapping_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, MAC=1, NAME=2, DESC=3)

		self.assertIn("DEBUG:ip4_mapping_invalid:10 records read", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_invalid:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_invalid:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_invalid:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:ip4_mapping_invalid:10 MAC mappings total", case_loggy.output)

		for invalid in (

			"IP4", 420, 69., True, False, None, ..., print, LibraryContrib,
			[ self.ip2 ],( self.ip2, ),{ self.ip2 },{ "IP4": self.ip2 }
		):
			with self.subTest(addr=invalid):

				self.assertIsNone(self.test_case.ip4map(invalid))
				self.assertIsNone(self.test_case.ip4map_mac(invalid))
				self.assertIsNone(self.test_case.ip4map_name(invalid))
				self.assertIsNone(self.test_case.ip4map_desc(invalid))








	def test_mac_mapping_valid(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="mac_mapping_valid",
				init_level=10,
			)
		)
		with self.assertLogs("mac_mapping_valid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, MAC=1, NAME=2, DESC=3)

		self.assertIn("DEBUG:mac_mapping_valid:10 records read", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_valid:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_valid:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_valid:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_valid:10 MAC mappings total", case_loggy.output)

		self.assertEqual(self.test_case.macmap_ip4(self.lmac0), self.ip0)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac1), self.ip1)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac2), self.ip2)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac3), self.ip3)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac4), self.ip4)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac5), self.ip5)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac6), self.ip6)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac7), self.ip7)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac8), self.ip8)
		self.assertEqual(self.test_case.macmap_ip4(self.lmac9), self.ip9)

		self.assertEqual(self.test_case.macmap_name(self.lmac0), self.name0)
		self.assertEqual(self.test_case.macmap_name(self.lmac1), self.name1)
		self.assertEqual(self.test_case.macmap_name(self.lmac2), self.name2)
		self.assertEqual(self.test_case.macmap_name(self.lmac3), self.name3)
		self.assertEqual(self.test_case.macmap_name(self.lmac4), self.name4)
		self.assertEqual(self.test_case.macmap_name(self.lmac5), self.name5)
		self.assertEqual(self.test_case.macmap_name(self.lmac6), self.name6)
		self.assertEqual(self.test_case.macmap_name(self.lmac7), self.name7)
		self.assertEqual(self.test_case.macmap_name(self.lmac8), self.name8)
		self.assertEqual(self.test_case.macmap_name(self.lmac9), self.name9)

		self.assertEqual(self.test_case.macmap_desc(self.lmac0), self.desc0)
		self.assertEqual(self.test_case.macmap_desc(self.lmac1), self.desc1)
		self.assertEqual(self.test_case.macmap_desc(self.lmac2), self.desc2)
		self.assertEqual(self.test_case.macmap_desc(self.lmac3), self.desc3)
		self.assertEqual(self.test_case.macmap_desc(self.lmac4), self.desc4)
		self.assertEqual(self.test_case.macmap_desc(self.lmac5), self.desc5)
		self.assertEqual(self.test_case.macmap_desc(self.lmac6), self.desc6)
		self.assertEqual(self.test_case.macmap_desc(self.lmac7), self.desc7)
		self.assertEqual(self.test_case.macmap_desc(self.lmac8), self.desc8)
		self.assertEqual(self.test_case.macmap_desc(self.lmac9), self.desc9)
		self.assertEqual(

			self.test_case.macmap(self.lmac0),
			{
				"IP4":	self.ip0,
				"NAME":	self.name0,
				"DESC":	self.desc0,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac1),
			{
				"IP4":	self.ip1,
				"NAME":	self.name1,
				"DESC":	self.desc1,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac2),
			{
				"IP4":	self.ip2,
				"NAME":	self.name2,
				"DESC":	self.desc2,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac3),
			{
				"IP4":	self.ip3,
				"NAME":	self.name3,
				"DESC":	self.desc3,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac4),
			{
				"IP4":	self.ip4,
				"NAME":	self.name4,
				"DESC":	self.desc4,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac5),
			{
				"IP4":	self.ip5,
				"NAME":	self.name5,
				"DESC":	self.desc5,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac6),
			{
				"IP4":	self.ip6,
				"NAME":	self.name6,
				"DESC":	self.desc6,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac7),
			{
				"IP4":	self.ip7,
				"NAME":	self.name7,
				"DESC":	self.desc7,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac8),
			{
				"IP4":	self.ip8,
				"NAME":	self.name8,
				"DESC":	self.desc8,
			}
		)
		self.assertEqual(

			self.test_case.macmap(self.lmac9),
			{
				"IP4":	self.ip9,
				"NAME":	self.name9,
				"DESC":	self.desc9,
			}
		)








	def test_mac_mapping_invalid(self):

		self.test_case = MaraudersMap(
			LibraryContrib(

				handler=self.MARAUDERS_HANDLER,
				init_name="mac_mapping_invalid",
				init_level=10,
			)
		)
		with self.assertLogs("mac_mapping_invalid", 10) as case_loggy:
			self.test_case.CSV(self.CSV_MAP_1, ";", IP4=0, MAC=1, NAME=2, DESC=3)

		self.assertIn("DEBUG:mac_mapping_invalid:10 records read", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_invalid:10 new IP4 mappings done", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_invalid:10 IP4 mappings total", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_invalid:10 new MAC mappings done", case_loggy.output)
		self.assertIn("DEBUG:mac_mapping_invalid:10 MAC mappings total", case_loggy.output)

		for invalid in (

			"MAC", 420, 69., True, False, None, ..., print, LibraryContrib,
			[ self.lmac4 ],( self.lmac4, ),{ self.lmac4 },{ "MAC": self.lmac4 }
		):
			with self.subTest(addr=invalid):

				self.assertIsNone(self.test_case.macmap(invalid))
				self.assertIsNone(self.test_case.macmap_ip4(invalid))
				self.assertIsNone(self.test_case.macmap_name(invalid))
				self.assertIsNone(self.test_case.macmap_desc(invalid))








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pygwarts.tests.hagrid				import MediumSet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.thrivables			import Copse
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.planting.weeds		import SprigTrimmer
from	pygwarts.hagrid.bloom.weeds			import Efflorescence
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class EdgeMultiEffloresce(MediumSet):

	"""
		Efflorescence only, with branches, with and without bough included
		Double sprout copse
		Double bough copse
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.MEDIUM_HANDLER_7): os.remove(cls.MEDIUM_HANDLER_7)

	@classmethod
	def setUpClass(cls):

		"""
			d- MEDIUM_SET_BOUGH_1
				f- lelik
				f- bolik
				d- bin
					f- untouchable
					f- mister
					d- in my bag
			d- MEDIUM_SET_BOUGH_2
				f- lelik
				f- bolik
				d- bin
					f- untouchable
					f- missis
					d- too bad

			d- MEDIUM_SET_SPROUT_1
				f- lelik
				d- bin
					f- untouchable
					f- mister
					d- in my bag
			d- MEDIUM_SET_SPROUT_2
				f- bolik
				d- bin
					f- untouchable
					f- missis
					d- too bad
		"""

		cls.make_loggy_file(cls, cls.MEDIUM_HANDLER_7)


		# First sprout
		cls.d_mss1_b				= os.path.join(cls.MEDIUM_SET_SPROUT_1, "bin")
		cls.f_mss1_lelik			= os.path.join(cls.MEDIUM_SET_SPROUT_1, "lelik")
		cls.f_mss1_b_untouchable	= os.path.join(cls.d_mss1_b, "untouchable")
		cls.f_mss1_b_mister			= os.path.join(cls.d_mss1_b, "mister")
		# Second sprout
		cls.d_mss2_b				= os.path.join(cls.MEDIUM_SET_SPROUT_2, "bin")
		cls.f_mss2_bolik			= os.path.join(cls.MEDIUM_SET_SPROUT_2, "bolik")
		cls.f_mss2_b_untouchable	= os.path.join(cls.d_mss2_b, "untouchable")
		cls.d_mss2_b_tb				= os.path.join(cls.d_mss2_b, "too bad")
		# First bough
		cls.tg1_f_lelik				= os.path.join(cls.MEDIUM_SET_BOUGH_1, "lelik")
		cls.tg1_f_bolik				= os.path.join(cls.MEDIUM_SET_BOUGH_1, "bolik")
		cls.tg1_d_b					= os.path.join(cls.MEDIUM_SET_BOUGH_1, "bin")
		cls.tg1_f_b_untouchable		= os.path.join(cls.tg1_d_b, "untouchable")
		cls.tg1_f_b_mister			= os.path.join(cls.tg1_d_b, "mister")
		cls.tg1_d_b_imb				= os.path.join(cls.tg1_d_b, "in my bag")
		# Sceond bough
		cls.tg2_f_lelik				= os.path.join(cls.MEDIUM_SET_BOUGH_2, "lelik")
		cls.tg2_f_bolik				= os.path.join(cls.MEDIUM_SET_BOUGH_2, "bolik")
		cls.tg2_d_b					= os.path.join(cls.MEDIUM_SET_BOUGH_2, "bin")
		cls.tg2_f_b_untouchable		= os.path.join(cls.tg2_d_b, "untouchable")
		cls.tg2_f_b_missis			= os.path.join(cls.tg2_d_b, "missis")
		cls.tg2_d_b_tb				= os.path.join(cls.tg2_d_b, "too bad")




	def setUp(self):
		self.clean()

		self.fmake(self.f_mss1_lelik, "bolik's brother")
		self.fmake(self.f_mss1_b_untouchable, "can't touch this")
		self.fmake(self.f_mss1_b_mister, "what is his name?")


		self.fmake(self.f_mss2_bolik, "lelik's brother")
		self.fmake(self.f_mss2_b_untouchable, "can't touch this")
		os.makedirs(self.d_mss2_b_tb, exist_ok=True)


		self.fmake(self.tg1_f_lelik, "bolik's brother")
		self.fmake(self.tg1_f_bolik, "lelik's brother")
		self.fmake(self.tg1_f_b_untouchable, "can't touch this")
		self.fmake(self.tg1_f_b_mister, "what is his name?")
		os.makedirs(self.tg1_d_b_imb, exist_ok=True)


		self.fmake(self.tg2_f_lelik, "bolik's brother")
		self.fmake(self.tg2_f_bolik, "lelik's brother")
		self.fmake(self.tg2_f_b_untouchable, "can't touch this")
		self.fmake(self.tg2_f_b_missis, "ain't real, actually, i guess")
		os.makedirs(self.tg2_d_b_tb, exist_ok=True)








	def test_distributed_effloresce(self):
		class Forest(Copse):

			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_7
				init_name	= "distributed_effloresce"
				init_level	= 10

			class First(Tree):

				bough = self.MEDIUM_SET_BOUGH_1
				class leafs(SiftingController):

					# include for growing/thriving and for trimming
					include = rf".+",
					# no bough root mess
					exclude = (

						( rf"{self.MEDIUM_SET_BOUGH_1}/[^/]+$", )
						if os.name == "posix" else
						( self.MEDIUM_SET_BOUGH_1.replace("\\","\\\\") + r"\\[^(\\)]+$", )
					)

				class twigs(SiftingController):

					# include for growing/thriving and for trimming
					include = rf".+",
					# no bough root mess
					exclude = (

						( rf"{self.MEDIUM_SET_BOUGH_1}/[^/]+$", )
						if os.name == "posix" else
						( self.MEDIUM_SET_BOUGH_1.replace("\\","\\\\") + r"\\[^(\\)]+$", )
					)

			class Second(Tree):

				bough = self.MEDIUM_SET_BOUGH_2
				class leafs(SiftingController):

					# include for growing/thriving and for trimming
					include = rf".+",
					# no bough root mess
					exclude = (

						( rf"{self.MEDIUM_SET_BOUGH_2}/[^/]+$", )
						if os.name == "posix" else
						( self.MEDIUM_SET_BOUGH_2.replace("\\","\\\\") + r"\\[^(\\)]+$", )
					)

				class twigs(SiftingController):

					# include for growing/thriving and for trimming
					include = rf".+",
					# no bough root mess
					exclude = (

						( rf"{self.MEDIUM_SET_BOUGH_2}/[^/]+$", )
						if os.name == "posix" else
						( self.MEDIUM_SET_BOUGH_2.replace("\\","\\\\") + r"\\[^(\\)]+$", )
					)



			class trim(SprigTrimmer):	pass
			class clean(Efflorescence):
				branches	= {

					self.MEDIUM_SET_BOUGH_1: ( self.MEDIUM_SET_SPROUT_1, ),
					self.MEDIUM_SET_BOUGH_2: ( self.MEDIUM_SET_SPROUT_2, ),
				}


			@fssprout(self.MEDIUM_SET_SPROUT_2)
			@fssprout(self.MEDIUM_SET_SPROUT_1)
			class sync(Flourish):		pass




		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_lelik))
		self.assertTrue(os.path.isdir(self.d_mss1_b))
		self.assertTrue(os.path.isfile(self.f_mss1_b_untouchable))
		self.assertTrue(os.path.isfile(self.f_mss1_b_mister))


		# Sceond sprout
		self.assertTrue(os.path.isfile(self.f_mss2_bolik))
		self.assertTrue(os.path.isdir(self.d_mss2_b))
		self.assertTrue(os.path.isfile(self.f_mss2_b_untouchable))
		self.assertTrue(os.path.isdir(self.d_mss2_b_tb))


		# First bough
		self.assertTrue(os.path.isfile(self.tg1_f_lelik))
		self.assertTrue(os.path.isfile(self.tg1_f_bolik))
		self.assertTrue(os.path.isdir(self.tg1_d_b))
		self.assertTrue(os.path.isfile(self.tg1_f_b_untouchable))
		self.assertTrue(os.path.isfile(self.tg1_f_b_mister))
		self.assertTrue(os.path.isdir(self.tg1_d_b_imb))


		# Sceond bough
		self.assertTrue(os.path.isfile(self.tg2_f_lelik))
		self.assertTrue(os.path.isfile(self.tg2_f_bolik))
		self.assertTrue(os.path.isdir(self.tg2_d_b))
		self.assertTrue(os.path.isfile(self.tg2_f_b_untouchable))
		self.assertTrue(os.path.isfile(self.tg2_f_b_missis))
		self.assertTrue(os.path.isdir(self.tg2_d_b_tb))




		with self.assertLogs("distributed_effloresce", 10) as case_loggy:

			self.test_case = Forest()
			self.test_case.sync()

		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed leaf \"{self.tg1_f_lelik}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed leaf \"{self.tg1_f_bolik}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed twig \"{self.tg1_d_b}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed leaf \"{self.tg1_f_b_untouchable}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed leaf \"{self.tg1_f_b_mister}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:distributed_effloresce:Trimmed twig \"{self.tg1_d_b_imb}\"",
			case_loggy.output
		)


		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed leaf \"{self.tg2_f_lelik}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed leaf \"{self.tg2_f_bolik}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed twig \"{self.tg2_d_b}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed twig \"{self.tg2_f_b_untouchable}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:distributed_effloresce:Trimmed leaf \"{self.tg2_f_b_missis}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:distributed_effloresce:Trimmed twig \"{self.tg2_d_b_tb}\"",
			case_loggy.output
		)




		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_lelik))
		self.assertTrue(os.path.isdir(self.d_mss1_b))
		self.assertTrue(os.path.isfile(self.f_mss1_b_untouchable))
		self.assertTrue(os.path.isfile(self.f_mss1_b_mister))


		# Sceond sprout
		self.assertTrue(os.path.isfile(self.f_mss2_bolik))
		self.assertTrue(os.path.isdir(self.d_mss2_b))
		self.assertTrue(os.path.isfile(self.f_mss2_b_untouchable))
		self.assertTrue(os.path.isdir(self.d_mss2_b_tb))


		# First bough
		self.assertTrue(os.path.isfile(self.tg1_f_lelik))
		self.assertTrue(os.path.isfile(self.tg1_f_bolik))
		self.assertTrue(os.path.isdir(self.tg1_d_b))
		self.assertTrue(os.path.isfile(self.tg1_f_b_untouchable))
		self.assertTrue(os.path.isfile(self.tg1_f_b_mister))
		self.assertFalse(os.path.isdir(self.tg1_d_b_imb))


		# Sceond bough
		self.assertTrue(os.path.isfile(self.tg2_f_lelik))
		self.assertTrue(os.path.isfile(self.tg2_f_bolik))
		self.assertTrue(os.path.isdir(self.tg2_d_b))
		self.assertTrue(os.path.isfile(self.tg2_f_b_untouchable))
		self.assertFalse(os.path.isfile(self.tg2_f_b_missis))
		self.assertTrue(os.path.isdir(self.tg2_d_b_tb))








	def test_bough_roots_effloresce(self):
		class Forest(Copse):

			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_7
				init_name	= "bough_roots_effloresce"
				init_level	= 10

			class First(Tree):

				bough = self.MEDIUM_SET_BOUGH_1
				class leafs(SiftingController):	include = rf".+",
				class twigs(SiftingController):	include = rf".+",

			class Second(Tree):

				bough = self.MEDIUM_SET_BOUGH_2
				class leafs(SiftingController):	include = rf".+",
				class twigs(SiftingController):	include = rf".+",


			class trim(SprigTrimmer):	pass
			class clean(Efflorescence):
				branches	= {

					self.MEDIUM_SET_BOUGH_1: ( self.MEDIUM_SET_SPROUT_1, ),
					self.MEDIUM_SET_BOUGH_2: ( self.MEDIUM_SET_SPROUT_2, ),
				}


			@fssprout(self.MEDIUM_SET_SPROUT_2)
			@fssprout(self.MEDIUM_SET_SPROUT_1)
			class sync(Flourish):		pass




		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_lelik))
		self.assertTrue(os.path.isdir(self.d_mss1_b))
		self.assertTrue(os.path.isfile(self.f_mss1_b_untouchable))
		self.assertTrue(os.path.isfile(self.f_mss1_b_mister))


		# Sceond sprout
		self.assertTrue(os.path.isfile(self.f_mss2_bolik))
		self.assertTrue(os.path.isdir(self.d_mss2_b))
		self.assertTrue(os.path.isfile(self.f_mss2_b_untouchable))
		self.assertTrue(os.path.isdir(self.d_mss2_b_tb))


		# First bough
		self.assertTrue(os.path.isfile(self.tg1_f_lelik))
		self.assertTrue(os.path.isfile(self.tg1_f_bolik))
		self.assertTrue(os.path.isdir(self.tg1_d_b))
		self.assertTrue(os.path.isfile(self.tg1_f_b_untouchable))
		self.assertTrue(os.path.isfile(self.tg1_f_b_mister))
		self.assertTrue(os.path.isdir(self.tg1_d_b_imb))


		# Sceond bough
		self.assertTrue(os.path.isfile(self.tg2_f_lelik))
		self.assertTrue(os.path.isfile(self.tg2_f_bolik))
		self.assertTrue(os.path.isdir(self.tg2_d_b))
		self.assertTrue(os.path.isfile(self.tg2_f_b_untouchable))
		self.assertTrue(os.path.isfile(self.tg2_f_b_missis))
		self.assertTrue(os.path.isdir(self.tg2_d_b_tb))




		with self.assertLogs("bough_roots_effloresce", 10) as case_loggy:

			self.test_case = Forest()
			self.test_case.sync()




		self.no_loggy_levels(case_loggy.output, 30,40,50)




		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed leaf \"{self.tg1_f_lelik}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:bough_roots_effloresce:Trimmed leaf \"{self.tg1_f_bolik}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed twig \"{self.tg1_d_b}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed leaf \"{self.tg1_f_b_untouchable}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed leaf \"{self.tg1_f_b_mister}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:bough_roots_effloresce:Trimmed twig \"{self.tg1_d_b_imb}\"",
			case_loggy.output
		)


		self.assertIn(

			f"INFO:bough_roots_effloresce:Trimmed leaf \"{self.tg2_f_lelik}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed leaf \"{self.tg2_f_bolik}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed twig \"{self.tg2_d_b}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed twig \"{self.tg2_f_b_untouchable}\"",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:bough_roots_effloresce:Trimmed leaf \"{self.tg2_f_b_missis}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:bough_roots_effloresce:Trimmed twig \"{self.tg2_d_b_tb}\"",
			case_loggy.output
		)




		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_lelik))
		self.assertTrue(os.path.isdir(self.d_mss1_b))
		self.assertTrue(os.path.isfile(self.f_mss1_b_untouchable))
		self.assertTrue(os.path.isfile(self.f_mss1_b_mister))


		# Sceond sprout
		self.assertTrue(os.path.isfile(self.f_mss2_bolik))
		self.assertTrue(os.path.isdir(self.d_mss2_b))
		self.assertTrue(os.path.isfile(self.f_mss2_b_untouchable))
		self.assertTrue(os.path.isdir(self.d_mss2_b_tb))


		# First bough
		self.assertTrue(os.path.isfile(self.tg1_f_lelik))
		self.assertFalse(os.path.isfile(self.tg1_f_bolik))
		self.assertTrue(os.path.isdir(self.tg1_d_b))
		self.assertTrue(os.path.isfile(self.tg1_f_b_untouchable))
		self.assertTrue(os.path.isfile(self.tg1_f_b_mister))
		self.assertFalse(os.path.isdir(self.tg1_d_b_imb))


		# Sceond bough
		self.assertFalse(os.path.isfile(self.tg2_f_lelik))
		self.assertTrue(os.path.isfile(self.tg2_f_bolik))
		self.assertTrue(os.path.isdir(self.tg2_d_b))
		self.assertTrue(os.path.isfile(self.tg2_f_b_untouchable))
		self.assertFalse(os.path.isfile(self.tg2_f_b_missis))
		self.assertTrue(os.path.isdir(self.tg2_d_b_tb))








if	__name__ == "__main__" : unittest.main(verbosity=2)










from socket									import socket
from socket									import AF_INET
from socket									import SOCK_STREAM
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.spells				import patronus
from pygwarts.filch.nettherin				import is_valid_ip4
from pygwarts.filch.transportclaw			import is_valid_port








class PortScanner(Transmutable):

	"""
		Transport layer class for scanning host's listening ports. Must accept positional arguments:
			target_ip	- IP4 address of target host to scan, as string;
			target_port	- integer in range 0-65535 which represents a port number to scan.
		Once "target_ip" and "target_port" are validated, the classic socket instance will be created
		to try establish connection with target. If connection will be closed with exit code 0 it must
		mean the port is opened and listening, so True will be returned. Any other exit code will be
		treated as port closed, so False will be returned. In any other situations, including Exception
		raise, corresponding logging will occur for identification and None will be returned.
	"""

	def __call__(self, target_ip :str, target_port :int) -> bool | None :


		if	is_valid_ip4(target_ip):
			if	is_valid_port(target_port):
				self.loggy.debug(f"Scanning {target_ip}:{target_port}")


				try:

					with socket(AF_INET, SOCK_STREAM) as connection:
						code = connection.connect_ex(( target_ip,target_port ))


						if		code == 0: self.loggy.info(f"{target_ip}:{target_port} listening")
						else:	self.loggy.debug(f"Socket write error {code}")
						return	not code


				except	Exception as E: self.loggy.debug(f"Scanner failed with {patronus(E)}")
			else:		self.loggy.debug(f"Invalid port \"{target_port}\"")
		else:			self.loggy.debug(f"Invalid target \"{target_ip}\"")










from pathlib		import Path
from pygwarts.tests	import PygwartsTestCase








class FilchTestCase(PygwartsTestCase):

	"""
		Super for filch testing
	"""

	FILCH_ROOT			= Path(PygwartsTestCase._PygwartsTestCase__CWD) /"filch"
	APPPUF_HANDLER		= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"filch" /"apppuf.loggy")
	TRNSPRTCLAW_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"filch" /"trnsprtclaw.loggy")
	NETTHERIN_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"filch" /"nettherin.loggy")
	LINKINDOR_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"filch" /"linkindor.loggy")
	MARAUDERS_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"filch" /"maramap.loggy")

	valid_MAC_chars	= "0123456789AaBbCcDdEeFf"
	valid_IP_chars	= list(range(256))










import	os
import	unittest
from	pathlib					import Path
from	pygwarts.irma.contrib	import LibraryContrib








class PygwartsTestCase(unittest.TestCase):

	"""
		Super class with instruments.
	"""


	maxDiff		= None
	clean_up	= False
	__CWD		= Path.home().joinpath("pygwarts-test-folder")


	LOGGY_LEVELS_MAPPER	= {

		20: "INFO",
		30: "WARNING",
		40: "ERROR",
		50: "CRITICAL",
	}


	def tearDown(self):

		"""
			Common "tearDown" that assumes that any child will operate over variable "test_case" that has
			a Logger object to be closed explicitly.
		"""

		if	hasattr(self, "test_case"):
			if	hasattr(self.test_case, "loggy"):
				if	isinstance(self.test_case.loggy, LibraryContrib):
					self.test_case.loggy.close()




	def make_loggy_file(self, path :str):

		"""
			As LibraryContrib creates handler files anyway, current method serves as a cleaner,
			cause all tests takes place in DEBUG mode and hence files becomes flooded.
		"""

		loggy_folder = os.path.dirname(path)

		if		not os.path.isdir(loggy_folder):	os.makedirs(loggy_folder)
		with	open(path, "w") as loggy_F :		pass




	def fmake(self, path :str ="", content :str ="alohomora"):

		"""
			For creation of required files.
			Also may serves as a touch.
			Argument "path" must be end file absolute path.
		"""

		folder	= os.path.dirname(path)

		if		not os.path.isdir(folder):	os.makedirs(folder)
		with	open(path, "w") as F:		F.write(content)




	def no_loggy_levels(self, loggy :LibraryContrib, *levels):

		"""
			Asserts the LibraryContrib object "loggy" has not emitted "levels" levels messages.
		"""

		self.assertTrue(isinstance(loggy, list))
		self.assertTrue(levels)


		_levels = list(
			filter(

				bool,
				[
					L if
					L in self.LOGGY_LEVELS_MAPPER.values() else
					self.LOGGY_LEVELS_MAPPER.get(L)
					for L in levels
				]
			)
		)


		self.assertTrue(len(_levels) == len(levels))
		self.assertTrue(set(_levels) <= set(self.LOGGY_LEVELS_MAPPER.values()))


		for record in loggy:
			for level in _levels:

				with self.subTest(loggy=record, level=level):
					self.assertFalse(record.startswith(level))










from typing									import Any
from typing									import Callable
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.spells				import patronus
from pygwarts.filch.nettherin				import is_valid_ip4








class HostPing(Transmutable):

	"""
		Network layer class for sending ICMP ping request and receiving pong response. Must accept
		positional arguments:
			target	- IP4 address of target host to be discovered, as string;
			pinger	- callable that must accept "target", along with any additional flags as **kwargs,
					and implement actual ping request constructing and sending and pong response processing.
		Once "target" validated and "pinger" assured callable, "pinger" will be invoked with "target" and
		all keyword arguments from current object __call__. If "pinger" call results anything that is not
		None, which means successful pong received, this pong will be returned. In any other cases,
		including Exception raise during "pinger" call, corresponding logging will occur for identification
		and None will be returned.
	"""

	def __call__(self, target :str, pinger :Callable[[Any],Any], **kwargs) -> Any :


		if	is_valid_ip4(target):
			if	callable(pinger):


				self.loggy.debug(f"Sending {target} ping request")
				for k,v in kwargs.items() : self.loggy.debug(f"Using {k}: {v}")


				try:

					if	(pong := pinger(target, **kwargs)) is not None:

						self.loggy.info(f"Received {target} pong response")
						return pong
					else:
						self.loggy.debug(f"No response from {target}")


				except	Exception as E:

					self.loggy.debug(f"Ping failed due to {patronus(E)}")
			else:	self.loggy.debug(f"Invalid pinger \"{pinger}\"")
		else:		self.loggy.debug(f"Invalid IP4 address \"{target}\"")










import	os
import	unittest
from	pathlib								import Path
from	shutil								import rmtree
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.hagrid				import HagridTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.planting.peels		import GrowingPeel
from	pygwarts.hagrid.planting.peels		import ThrivingPeel








class PeelCases(HagridTestCase):

	"""
		Try out peels variants
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.PEEL_HANDLER): os.remove(cls.PEEL_HANDLER)

	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.PEEL_HANDLER)

		cls.peel_2_2 = os.path.join("not so deep", "almost deep")
		cls.peel_2_3 = os.path.join(cls.peel_2_2, "the deep", "very deep", "oh my deep")

		cls.d2peel_1 = os.path.join(cls.INIT_SET_SPROUT, "d2peel_1")
		cls.d2peel_2 = os.path.join(cls.INIT_SET_SPROUT, cls.peel_2_2, "d2peel_2")
		cls.d2peel_3 = os.path.join(cls.INIT_SET_SPROUT, cls.peel_2_3, "d2peel_3")
		cls.f2peel_1 = os.path.join(cls.INIT_SET_SPROUT, "f2peel_1")
		cls.f2peel_2 = os.path.join(cls.INIT_SET_SPROUT, cls.peel_2_2, "f2peel_2")
		cls.f2peel_3 = os.path.join(cls.INIT_SET_SPROUT, cls.peel_2_3, "f2peel_3")




	def test_growing_peel(self):
		class Sakura(Tree):

			bough = self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.PEEL_HANDLER
				init_name	= "growing_peel"
				init_level	= 10


			@GrowingPeel
			class grow(Transmutable):
				def __call__(*args) : pass


		with self.assertLogs("growing_peel", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_1)),
				Path(os.path.basename(self.d2peel_1)),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:growing_peel:Peeled growing bough \"{self.INIT_SET_BOUGH_1}\"",
				case_loggy.output
			)

			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_2)),
				Path(os.path.basename(self.d2peel_2)),
				Path(self.INIT_SET_BOUGH_1)
			)
			d2T = os.path.join(self.INIT_SET_BOUGH_1, self.peel_2_2)
			self.assertIn(f"DEBUG:growing_peel:Peeled growing bough \"{d2T}\"", case_loggy.output)

			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_3)),
				Path(os.path.basename(self.d2peel_3)),
				Path(self.INIT_SET_BOUGH_1)
			)
			d3T = os.path.join(self.INIT_SET_BOUGH_1, self.peel_2_3)
			self.assertIn(f"DEBUG:growing_peel:Peeled growing bough \"{d3T}\"", case_loggy.output)

			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_3)),
				Path(os.path.basename(self.d2peel_3)),
				Path(os.path.dirname(self.d2peel_3))
			)
			self.assertIn(f"DEBUG:growing_peel:Peeled growing bough \"{d3T}\"", case_loggy.output)








	def test_thriving_peel(self):
		testing_thrivings = "to", "the", "top",

		class Sakura(Tree):

			bough = self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.PEEL_HANDLER
				init_name	= "thriving_peel"
				init_level	= 10

			@ThrivingPeel(*testing_thrivings)
			class top_grow(Transmutable):
				def __call__(*args):	pass

			@ThrivingPeel(*testing_thrivings, to_peak=False)
			class bot_grow(Transmutable):
				def __call__(*args):	pass


		top_grown	= os.path.join(self.INIT_SET_BOUGH_1, *testing_thrivings)
		bot_grown	= os.path.join(self.INIT_SET_BOUGH_1, *testing_thrivings)
		top_d2T		= os.path.join(self.INIT_SET_BOUGH_1, self.peel_2_2, *testing_thrivings)
		top_d3T		= os.path.join(self.INIT_SET_BOUGH_1, self.peel_2_3, *testing_thrivings)
		bot_d2T		= os.path.join(self.INIT_SET_BOUGH_1, *testing_thrivings, self.peel_2_2)
		bot_d3T		= os.path.join(self.INIT_SET_BOUGH_1, *testing_thrivings, self.peel_2_3)
		self.test_case = Sakura()

		with self.assertLogs("thriving_peel", 10) as case_loggy:

			self.test_case.top_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_1)),
				Path(os.path.basename(self.d2peel_1)),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{top_grown}\"", case_loggy.output)

			self.test_case.top_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_2)),
				Path(os.path.basename(self.d2peel_2)),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{top_d2T}\"", case_loggy.output)

			self.test_case.top_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_3)),
				Path(os.path.basename(self.d2peel_3)),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{top_d3T}\"", case_loggy.output)

			self.test_case.top_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_3)),
				Path(os.path.basename(self.d2peel_3)),
				Path(os.path.dirname(self.d2peel_3))
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{top_d3T}\"", case_loggy.output)




			self.test_case.bot_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_1)),
				Path(os.path.basename(self.d2peel_1)),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{bot_grown}\"", case_loggy.output)

			self.test_case.bot_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_2)),
				Path(os.path.basename(self.d2peel_2)),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{bot_d2T}\"", case_loggy.output)

			self.test_case.bot_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_3)),
				Path(os.path.basename(self.d2peel_3)),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{bot_d3T}\"", case_loggy.output)

			self.test_case.bot_grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(os.path.dirname(self.d2peel_3)),
				Path(os.path.basename(self.d2peel_3)),
				Path(os.path.dirname(self.d2peel_3))
			)
			self.assertIn(f"DEBUG:thriving_peel:Peeled thriving bough \"{bot_d3T}\"", case_loggy.output)


		try:

			class SakuraRaise(Tree):

				bough = self.INIT_SET_BOUGH_1
				@ThrivingPeel("raise", "cause", False)
				class raising_grow(Transmutable):
					def __call__(*args):	pass


		except	Exception as E:

			self.assertIsInstance(E, AssertionError)
			self.assertEqual(str(E), f"Thriving must be with strings, not \"{type(False)}\"")








if	__name__ == "__main__" : unittest.main(verbosity=2)







